# keys for stacked lightcurve processing
MEAN = "mean"
MEDIAN = "median"
RMS = "rms"
UPPER_LIMIT = "ul"
NPOINTS = "npoints"
ZEROPOINT_EXT = "zeropoint"
FLUX_DENSITY_EXT = "fluxdensity"
KSTEST_NORM_EXT = "kstest_norm_pvalue"

# keys of the single exposure lightcurve
FLUX_EXT = "flux"
MAG_EXT = "mpro"
ERROR_EXT = "sig"
